/* sithwm - Minimalist Window Manager for X
 * see README for Copyright, license and other details. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "sithwm.h"
#include <sys/select.h>
#include <unistd.h>

static int enter_focus_countdown;
unsigned int last_current_info;

static void cleanup_clients(void)
{
#ifdef DEBUG
   int no_cli=0,no_del=0;
#endif
   Client *old_h = head_client;
   char bc_to_set = 'f'-COL_BASE;
   for (Client **p = &head_client;*p;) {
      Client*z=*p;

      if (z->dyn.remove) {
	 *p = z->next;
#ifdef DEBUG
	 LOG_DEBUG("\tremove: 0x%lx\n", z->dyn.window);
	 no_del++;
#endif
	 if (z->name != default_name)
	    XFree(z->name);
	 free(z);
      } else {
#ifdef DEBUG
	 no_cli++;
#endif
	 if (bc_to_set != z->dyn.border_color) {
	    XSetWindowBorder(dpy, z->dyn.window, z->screen->pixel[(int)bc_to_set]);
	    z->dyn.border_color = bc_to_set;
	    enter_focus_countdown=12;
	 }
         bc_to_set = 'e'-COL_BASE;
         p = &z->next;
      }
   }

   if (old_h != head_client)
      select_client(head_client, XACT_DELETED<<FUNC_SHIFT);

#ifdef DEBUG
   if (no_del)
      LOG_DEBUG("\tcleanup_clients() : window count now %d\n", no_cli);
#endif
}

int solid_drag = 1;

struct Dragsweep dragsweep;

#ifdef SNAP
static int absmin(int a, int b) {
   if (abs(a) < abs(b))
      return a;
   return b;
}


static void snap_client(Client *c)
{
   int dx = opt_snap, dy = opt_snap;
   struct area cc;
   cc.w = c->dyn.area.w;
   cc.h = c->dyn.area.h;
   toviewpos(c, &cc.pos);

   /* snap to screen border */
   dx = absmin(dx, -cc.pos.x);
   dy = absmin(dy, -cc.pos.y);
   dx = absmin(dx, c->screen->area.w - cc.pos.x - cc.w);
   dy = absmin(dy, c->screen->area.h - cc.pos.y - cc.h);

#ifdef SNAP_TO_WINDOWS
   Client *ci;
   for (ci = head_client; ci; ci = ci->next) {
      if (ci != c && ci->screen == c->screen) {
	 struct pos cip;
	 toviewpos(ci, &cip);
	 if (cip.y - cc.h - cc.pos.y <= opt_snap && cc.pos.y - ci->dyn.area.h - cip.y <= opt_snap) {
	    dx = absmin(dx, cip.x + ci->dyn.area.w - cc.pos.x);
	    dx = absmin(dx, cip.x + ci->dyn.area.w - cc.pos.x - cc.w);
	    dx = absmin(dx, cip.x - cc.pos.x - cc.w);
	    dx = absmin(dx, cip.x - cc.pos.x);
	 }
	 if (cip.x - cc.w - cc.pos.x <= opt_snap && cc.pos.x - ci->dyn.area.w - cip.x <= opt_snap) {
	    dy = absmin(dy, cip.y + ci->dyn.area.h - cc.pos.y);
	    dy = absmin(dy, cip.y + ci->dyn.area.h - cc.pos.y - cc.h);
	    dy = absmin(dy, cip.y - cc.pos.y - cc.h);
	    dy = absmin(dy, cip.y - cc.pos.y);
	 }
      }
   }
#endif

   if (abs(dx) < opt_snap)
      cc.pos.x += dx;
   if (abs(dy) < opt_snap)
      cc.pos.y += dy;
   toworldpos(c, cc.pos.x, cc.pos.y, &c->dyn.area.pos);
}
#endif /* def SNAP */

unsigned int infow_scaler;

void event_main_loop(void)
{
   static unsigned int infow_state_counter;
   static unsigned int infow_post_counter;
   static unsigned int has_waited_counter;
   unsigned int thewt;
   XEvent zevent;

   /* main event loop here */
   for (;;) {
      while (XCheckMaskEvent(dpy, -1, &zevent)) {
	 ScreenInfo*screen;
	 Client *client = find_client(zevent.xany.window);
	 for (ScreenInfo*s2 = screen = screens; s2 < screens+num_screens; s2++) {
	    if (s2->infow == zevent.xany.window &&
		s2->infow_func<INFOW_MAP &&
		infow_post_counter!=has_waited_counter)
	    {
	       LOG_DEBUG("\tunmap capt, because event\n");
	       XUnmapWindow(dpy, zevent.xany.window);
	       s2->infow_func = INFOW_OFF;
	    }
	    if (s2->root == zevent.xkey.root || s2->root == zevent.xany.window)
	       screen = s2;
	 }
	 switch (zevent.type) {
	 case KeyPress:
	    LOG_DEBUG("KeyPress %lx +%d+%d\n", zevent.xkey.window, zevent.xkey.x, zevent.xkey.y);
	    infow_post_counter = has_waited_counter;
#ifdef MENU
	    if (screen->infow_func == INFOW_MENU)
	       do_menu(&zevent);
	    else
#endif
	    {
	       KeySym key = XKeycodeToKeysym(dpy,zevent.xkey.keycode,0);
	       do_func(&zevent,key);
	    }
	    break;
	 case ButtonPress:
	    LOG_DEBUG("ButtonPress %lx +%d+%d\n", zevent.xbutton.window, zevent.xbutton.x, zevent.xbutton.y);
	    do_func(&zevent, XK_Pointer_Button1+zevent.xbutton.button - Button1);
	    break;
	 case MotionNotify:
	    LOG_DEBUG("MotionNotify %lx +%d+%d \n", zevent.xmotion.window, zevent.xmotion.x, zevent.xmotion.y);
	    client = find_client(dragsweep.w);
	    if (client) {
	       int ox = zevent.xmotion.x + dragsweep.pos.x;
	       int oy = zevent.xmotion.y + dragsweep.pos.y;
	       if (dragsweep.fl) {
		  ox -= client->dyn.area.pos.x;
		  oy -= client->dyn.area.pos.y;
		  if (ox<22) ox=22;
		  if (oy<22) oy=22;
		  client->dyn.area.w = ox;
		  client->dyn.area.h = oy;
	       } else {
		  client->dyn.area.pos.x = ox;
		  client->dyn.area.pos.y = oy;
#ifdef SNAP
		  if (opt_snap)
		     snap_client(client);
#endif
	       }

	       if (solid_drag)
		  select_client(client, FMOD_MOVE_RESIZE);
	       break;
	    }
	 case ButtonRelease:
	    XUngrabPointer(dpy, CurrentTime);
            select_window(dragsweep.w, FMOD_MOVE_RESIZE);
	    dragsweep.w = 0;
	    break;
	 case ConfigureRequest: {
	    XWindowChanges wc;
	    //wc.sibling = zevent.xconfigurerequest.above;
	    //wc.stack_mode = zevent.xconfigurerequest.detail;
	    wc.width = zevent.xconfigurerequest.width;
	    wc.height = zevent.xconfigurerequest.height;
	    wc.x = zevent.xconfigurerequest.x;
	    wc.y = zevent.xconfigurerequest.y;
	    //
	    // Mask away stack mode for menus in opera to work.
	    // (this fails in aewm and evilwm also ... )
	    // xterm and gkrellm needs to have ther size-config's
	    // returned.
	    // Other clients may need StackMode :-(
	    // 
	    XConfigureWindow(dpy, zevent.xconfigurerequest.window,
			     zevent.xconfigurerequest.value_mask&(~(CWStackMode|CWSibling)),
			     &wc);
	    LOG_DEBUG("handle_configure_request() : 0x%lx configured to %dx%d+%d+%d\n",
		      zevent.xconfigurerequest.window, wc.width, wc.height, wc.x, wc.y);
	 }break;
	 case ConfigureNotify:
	    if (zevent.xconfigure.window != screen->infow)
	    {
	       client = find_client(zevent.xconfigurerequest.window);
	       if (client) {
		  client->dyn.area.w = zevent.xconfigure.width;
		  client->dyn.area.h = zevent.xconfigure.height;
		  toworldpos(client, zevent.xconfigure.x,  zevent.xconfigure.y, &client->dyn.area.pos);
	       }
	    }
	    LOG_DEBUG("ConfigureNotify: 0x%lx configured to %dx%d+%d+%d\n", zevent.xconfigure.window,
		      zevent.xconfigure.width, zevent.xconfigure.height, zevent.xconfigure.x, zevent.xconfigure.y);
	    break;
	 case MapRequest:
	    if (!select_window(zevent.xmaprequest.window, FMOD_MOVE_RESIZE|FMOD_RAISE|FMOD_FOCUS))
	       make_new_client(zevent.xmaprequest.window, screen, 0);
	    break;
	 case ColormapNotify:
	    if (client && zevent.xcolormap.new) {
		client->cmap = zevent.xcolormap.colormap;
		XInstallColormap(dpy, client->cmap);
	    }
	    break;
	 case FocusIn:
	 case  FocusOut:
	    LOG_DEBUG("Focus: 0x%lx %s\n", zevent.xfocus.window,zevent.xfocus.type==FocusIn?"In":"Out");
            break;
	 case CreateNotify:
	    LOG_DEBUG("CreateNotify: 0x%lx\n", zevent.xcreatewindow.window);
	    break;
	 case  DestroyNotify:
	    LOG_DEBUG("DestroyNotify: 0x%lx\n", zevent.xdestroywindow.window);
	    break;
	 case  MapNotify:
	    LOG_DEBUG("MapNotify: 0x%lx\n", zevent.xmap.window);
	    break;
	 case UnmapNotify:
	    LOG_DEBUG("UnmapNotify: 0x%lx\n", zevent.xunmap.window);
	    client = find_client(zevent.xunmap.window);
	    if (client)
	       client->dyn.remove = 1;
	 break;
	 case EnterNotify:
	    LOG_DEBUG("EnterNotify: 0x%lx \n", zevent.xcrossing.window);
	    if (zevent.xcrossing.window == zevent.xcrossing.root)
	       current_screen = screen;
	    if (zevent.xcrossing.window != zevent.xcrossing.root)
	    if (screen->infow_func!=INFOW_MENU)
	       select_window(zevent.xcrossing.window, (XACT_ENTER<<FUNC_SHIFT)|FMOD_FOCUS);
	    break;
	 case PropertyNotify:
	    LOG_DEBUG("PropertyNotify: 0x%lx \n", zevent.xproperty.window);
	    if (client) {
	       client->dyn.named = 1;
	       get_client_props(client);
	    }
	    last_current_info = 0;
	    break;
	 default:
	    LOG_DEBUG("default: 0x%lx %i\n", zevent.xany.window, zevent.type);
	 }
      }

      int rx,ry,wx,wy;
      unsigned int mask;
      Window wns[3];
      wns[0] = current_screen->root;
      for (int i=0;i<2;i++)
	 XQueryPointer(dpy, wns[i], &wns[0], &wns[i+1],
		    &rx,&ry, &wx,&wy, &mask);
      
#define WTSHIFT (7)
#define SECTOWT(N) ((unsigned int)(N)*(1<<WTSHIFT))

      cleanup_clients();

      Client *lhclient = head_client;

      for (ScreenInfo*s2=screens+num_screens-1 ; s2>= screens ; s2--) {
	 if (((has_waited_counter-infow_post_counter)>SECTOWT(7)) ||
	     (s2->infow_func == INFOW_MAP && (mask&modifiers[0].mask) == 0) ||
	     (s2->infow_func == INFOW_CAPTION && (!lhclient || !lhclient->dyn.named))
	     )
	 {
	    if (s2->infow_func == INFOW_MAP) {
	       if (!select_window(s2->hilit, s2->hilit_funcmask))
		  setview();
	       s2->hilit = 0;
	       s2->do_infow_func = 0;
	    }
	    XUnmapWindow(dpy, s2->infow);
	    s2->infow_func = 0;
	 }
      }

      if (current_screen->infow_func != INFOW_MAP) {
	 infow_state_counter = 0;
	 infow_scaler = 0;
      } else {
	 infow_state_counter ++;
	 int scaler = infow_state_counter>opt_popup_time_in;
	 if (scaler != infow_scaler)
	    current_screen->do_infow_func = INFOW_MAP;
	 infow_scaler = scaler;
      }

      if (current_screen->do_infow_func) {
         int zz = current_screen->do_infow_func;
	 current_screen->do_infow_func = 0;
	 update_info_window(zz);
	 infow_post_counter = has_waited_counter;
      }
      
      lhclient = head_client;

      {
	 Window to_focus = (current_screen->infow_func > INFOW_CAPTION)?
	    current_screen->infow
	    : (lhclient?lhclient->dyn.window:0);
	 if (to_focus)
	    XSetInputFocus(dpy, to_focus, RevertToPointerRoot, CurrentTime);
      }

      if (lhclient && lhclient->dyn.named && lhclient->screen->infow_func<=INFOW_CAPTION) {
	 unsigned int nhash = hash_val(&lhclient->dyn, sizeof lhclient->dyn, 11);
	 if (nhash != last_current_info)
	 {
	    LOG_DEBUG("Caption: 0x%lx \n", lhclient->dyn.window);
	    lhclient->screen->do_infow_func = INFOW_CAPTION;
	 }
	 last_current_info = nhash;
      }

      if (mask&modifiers[0].mask)
 	 thewt = 0;
      thewt++;
      struct timeval ttxx;
      ttxx.tv_usec=(thewt%(1<<WTSHIFT))*(1000000>>WTSHIFT);
      ttxx.tv_sec=thewt>>WTSHIFT;

      XFlush(dpy);
      int x_connect_fd = ConnectionNumber(dpy);
      fd_set fdset;
      FD_ZERO(&fdset);
      FD_SET(x_connect_fd, &fdset);

      if ( dragsweep.w || enter_focus_countdown>0 ) {
	 Client *dc = find_client(dragsweep.w);
	 if (enter_focus_countdown>0) {
	    enter_focus_countdown--;
	    dc = lhclient;
	 }
	 thewt = 0;

	 if (dc) {
	    struct pos r;
	    ScreenInfo*screen = dc->screen;
	    toviewpos(dc, &r);
	    r.x += opt_bw-1;
	    r.y += opt_bw-1;
	    glob_gcv.function = GXinvert;
	    glob_gcv.dash_offset = (has_waited_counter++);
	    glob_gcv.dashes = 16;
	    glob_gcv.line_style = LineOnOffDash;
	    XChangeGC(dpy, screen->gc['d'-COL_BASE], GCFunction|GCLineStyle|GCDashOffset|GCDashList, &glob_gcv);
	    XGrabServer(dpy);
	    XDrawRectangle(dpy, screen->root, screen->gc['d'-COL_BASE],
			   r.x, r.y, dc->dyn.area.w+1, dc->dyn.area.h+1);
	    XFlush(dpy);
	    // Put in if drag rect will not show up:
	    // select(x_connect_fd + 1, &fdset, 0, 0, &ttxx);
	    XDrawRectangle(dpy, screen->root, screen->gc['d'-COL_BASE],
			   r.x, r.y, dc->dyn.area.w+1, dc->dyn.area.h+1);
	    XUngrabServer(dpy);
	    glob_gcv.function = GXcopy;
	    glob_gcv.line_style = LineSolid;
	    XChangeGC(dpy, screen->gc['d'-COL_BASE], GCFunction|GCLineStyle, &glob_gcv);
	 }
      }

      static unsigned int save_mask;
      static int  save_rx;
      if (select(x_connect_fd + 1, &fdset, 0, 0, &ttxx) || (rx != save_rx || save_mask != mask)) {
	 save_rx = rx;
	 save_mask = mask;
	 thewt = 0;
      } else {
	 if (current_screen->infow_func == INFOW_MAP && stacker_offset)
	    current_screen->do_infow_func = INFOW_MAP;
	 has_waited_counter += thewt;
	 if (lhclient) {
	    if (lhclient->dyn.window == wns[1]) {
	       lhclient->last_screen = lhclient->screen->area.pos;
	       lhclient->last_cursor.x = wx;
	       lhclient->last_cursor.y = wy;
	    }
	    if ( thewt>18) {
	       if ( memcmp(&lhclient->hist[lhclient->hist_index], &lhclient->dyn.area, sizeof(lhclient->dyn.area)) ) {
		  lhclient->hist_index = (lhclient->hist_index+1)%HISTORY;
		  lhclient->hist[lhclient->hist_index] = lhclient->dyn.area;
	       }
	       log_window_event("P", lhclient);
	       lhclient->dyn.named = 1;
	    }
	 }
	 if ( thewt>8 &&
	      memcmp(&current_screen->hist[current_screen->hist_index], &current_screen->area.pos, sizeof(current_screen->area.pos)) ) {
	    current_screen->hist_index = (current_screen->hist_index+1)%VIEW_HISTORY;
	    current_screen->hist[current_screen->hist_index] = current_screen->area.pos;
	 }
	 if (thewt>230) {
	    system(lock_command);
	    thewt = 0;
	 }
      }
   }
}
